=== modified file 'NEWS'
--- NEWS	2006-10-04 07:42:28 +0000
+++ NEWS	2006-10-05 07:34:58 +0000
@@ -1,6 +1,7 @@
 IN DEVELOPMENT
 
   IMPROVEMENTS:
+
     * ``bzr help commands`` output is now shorter (Aaron Bentley)
 
   INTERNALS:
@@ -13,6 +14,13 @@
       pass in inventory.ROOT_ID if you want the default root id value.
       (Robert Collins, John Arbash Meinel)
 
+    * New method ``WorkingTree.flush()`` which will write the current memory
+      inventory out to disk. At the same time, read_working_inventory will
+      no longer trash the current tree inventory if it has been modified within
+      the current lock, and the tree will now ``flush()`` automatically on
+      ``unlock()``. ``WorkingTree.set_root_id()`` has been updated to take
+      advantage of this functionality. (Robert Collins, John Arbash Meinel)
+
   BUG FIXES:
 
     * bzr branch/pull no longer complain about progress bar cleanup when

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-09-27 20:27:46 +0000
+++ bzrlib/errors.py	2006-10-05 07:34:58 +0000
@@ -183,6 +183,14 @@
         self.branch = branch
 
 
+class InventoryModified(BzrNewError):
+    """The current inventory for the tree %(tree)r has been modified, so a clean inventory cannot be read without data loss."""
+
+    def __init__(self, tree):
+        BzrNewError.__init__(self)
+        self.tree = tree
+
+
 class NoSuchId(BzrNewError):
     """The file id %(file_id)s is not present in the tree %(tree)s."""
     
@@ -212,6 +220,14 @@
         self.url = url
 
 
+class NotWriteLocked(BzrNewError):
+    """%(not_locked)r is not write locked but needs to be."""
+
+    def __init__(self, not_locked):
+        BzrNewError.__init__(self)
+        self.not_locked = not_locked
+
+
 class BzrCommandError(BzrNewError):
     """Error from user command"""
 

=== modified file 'bzrlib/tests/test_errors.py'
--- bzrlib/tests/test_errors.py	2006-09-27 19:20:35 +0000
+++ bzrlib/tests/test_errors.py	2006-10-05 07:34:58 +0000
@@ -26,6 +26,13 @@
 
 class TestErrors(TestCaseWithTransport):
 
+    def test_inventory_modified(self):
+        error = errors.InventoryModified("a tree to be repred")
+        self.assertEqualDiff("The current inventory for the tree 'a tree to "
+            "be repred' has been modified, so a clean inventory cannot be "
+            "read without data loss.",
+            str(error))
+
     def test_no_repo(self):
         dir = bzrdir.BzrDir.create(self.get_url())
         error = errors.NoRepositoryPresent(dir)
@@ -38,6 +45,12 @@
             "atree.",
             str(error))
 
+    def test_not_write_locked(self):
+        error = errors.NotWriteLocked('a thing to repr')
+        self.assertEqualDiff("'a thing to repr' is not write locked but needs "
+            "to be.",
+            str(error))
+
     def test_up_to_date(self):
         error = errors.UpToDateFormat(bzrdir.BzrDirFormat4())
         self.assertEqualDiff("The branch format Bazaar-NG branch, "

=== modified file 'bzrlib/tests/workingtree_implementations/__init__.py'
--- bzrlib/tests/workingtree_implementations/__init__.py	2006-10-04 07:42:28 +0000
+++ bzrlib/tests/workingtree_implementations/__init__.py	2006-10-05 07:34:58 +0000
@@ -56,6 +56,7 @@
         'bzrlib.tests.workingtree_implementations.test_changes_from',
         'bzrlib.tests.workingtree_implementations.test_commit',
         'bzrlib.tests.workingtree_implementations.test_executable',
+        'bzrlib.tests.workingtree_implementations.test_flush',
         'bzrlib.tests.workingtree_implementations.test_get_parent_ids',
         'bzrlib.tests.workingtree_implementations.test_is_control_filename',
         'bzrlib.tests.workingtree_implementations.test_is_ignored',
@@ -65,6 +66,7 @@
         'bzrlib.tests.workingtree_implementations.test_pull',
         'bzrlib.tests.workingtree_implementations.test_put_file',
         'bzrlib.tests.workingtree_implementations.test_mkdir',
+        'bzrlib.tests.workingtree_implementations.test_read_working_inventory',
         'bzrlib.tests.workingtree_implementations.test_set_root_id',
         'bzrlib.tests.workingtree_implementations.test_unversion',
         'bzrlib.tests.workingtree_implementations.test_workingtree',

=== added file 'bzrlib/tests/workingtree_implementations/test_flush.py'
--- bzrlib/tests/workingtree_implementations/test_flush.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/workingtree_implementations/test_flush.py	2006-10-05 07:34:58 +0000
@@ -0,0 +1,68 @@
+# (C) 2006 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for WorkingTree.flush."""
+
+from bzrlib import errors, inventory
+from bzrlib.tests.workingtree_implementations import TestCaseWithWorkingTree
+
+
+class TestFlush(TestCaseWithWorkingTree):
+
+    def test_flush_fresh_tree(self):
+        tree = self.make_branch_and_tree('t1')
+        tree.lock_write()
+        try:
+            tree.flush()
+        finally:
+            tree.unlock()
+
+    def test_flush_when_inventory_is_modified(self):
+        # when doing a flush the inventory should be written if needed.
+        # we test that by changing the inventory (using
+        # _set_inventory for now until add etc have lazy writes of
+        # the inventory on unlock).
+        tree = self.make_branch_and_tree('tree')
+        # prepare for a series of changes that will modify the 
+        # inventory
+        tree.lock_write()
+        try:
+            old_root = tree.get_root_id()
+            tree.set_root_id('new-root')
+            # to detect that the inventory is written by flush, we
+            # first check that it was not written yet.
+            reference_tree = tree.bzrdir.open_workingtree()
+            self.assertEqual(old_root, reference_tree.get_root_id())
+            # now flush the tree which should write the inventory.
+            tree.flush()
+            # and check it was written using another reference tree
+            reference_tree = tree.bzrdir.open_workingtree()
+            self.assertEqual('new-root', reference_tree.get_root_id())
+        finally:
+            tree.unlock()
+            
+    def test_flush_with_read_lock_fails(self):
+        """Flush cannot be used during a read lock."""
+        tree = self.make_branch_and_tree('t1')
+        tree.lock_read()
+        try:
+            self.assertRaises(errors.NotWriteLocked, tree.flush)
+        finally:
+            tree.unlock()
+            
+    def test_flush_with_no_lock_fails(self):
+        tree = self.make_branch_and_tree('t1')
+        self.assertRaises(errors.NotWriteLocked, tree.flush)

=== modified file 'bzrlib/tests/workingtree_implementations/test_locking.py'
--- bzrlib/tests/workingtree_implementations/test_locking.py	2006-09-27 22:12:53 +0000
+++ bzrlib/tests/workingtree_implementations/test_locking.py	2006-10-05 07:34:58 +0000
@@ -100,6 +100,52 @@
         self.assertTrue(wt.branch.is_locked())
         wt.branch.unlock()
         
+    def _test_unlock_with_lock_method(self, methodname):
+        """Create a tree and then test its unlocking behaviour.
+
+        :param methodname: The lock method to use to establish locks.
+        """
+        # when unlocking the last lock count from tree_write_lock,
+        # the tree should do a flush().
+        # we test that by changing the inventory using set_root_id
+        tree = self.make_branch_and_tree('tree')
+        # prepare for a series of changes that will modify the 
+        # inventory
+        getattr(tree, methodname)()
+        # note that we dont have a try:finally here because of two reasons:
+        # firstly there will only be errors reported if the test fails, and 
+        # when it fails thats ok as long as the test suite cleanup still works,
+        # which it will as the lock objects are released (thats wher ethe 
+        # warning comes from.  Secondly, its hard in this test to be 
+        # sure that we've got the right interactions between try:finally
+        # and the lock/unlocks we are doing.
+        getattr(tree, methodname)()
+        # this should really do something within the public api
+        # e.g. mkdir('foo') but all the mutating methods at the
+        # moment trigger inventory writes and thus will not 
+        # let us trigger a read-when-dirty situation.
+        old_root = tree.get_root_id()
+        tree.set_root_id('new-root')
+        # to detect that the inventory is written by unlock, we
+        # first check that it was not written yet.
+        reference_tree = tree.bzrdir.open_workingtree()
+        self.assertEqual(old_root, reference_tree.get_root_id())
+        # now unlock the second held lock, which should do nothing.
+        tree.unlock()
+        reference_tree = tree.bzrdir.open_workingtree()
+        self.assertEqual(old_root, reference_tree.get_root_id())
+        # unlocking the first lock we took will now flush.
+        tree.unlock()
+        # and check it was written using another reference tree
+        reference_tree = tree.bzrdir.open_workingtree()
+        self.assertEqual('new-root', reference_tree.get_root_id())
+
+    def test_unlock_from_tree_write_lock_flushes(self):
+        self._test_unlock_with_lock_method("lock_tree_write")
+        
+    def test_unlock_from_write_lock_flushes(self):
+        self._test_unlock_with_lock_method("lock_write")
+        
     def test_unlock_branch_failures(self):
         """If the branch unlock fails the tree must still unlock."""
         # The public interface for WorkingTree requires a branch, but

=== added file 'bzrlib/tests/workingtree_implementations/test_read_working_inventory.py'
--- bzrlib/tests/workingtree_implementations/test_read_working_inventory.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/workingtree_implementations/test_read_working_inventory.py	2006-10-05 07:34:58 +0000
@@ -0,0 +1,40 @@
+# (C) 2006 Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for WorkingTree.read_working_inventory."""
+
+from bzrlib import errors, inventory
+from bzrlib.tests.workingtree_implementations import TestCaseWithWorkingTree
+
+
+class TestReadWorkingInventory(TestCaseWithWorkingTree):
+
+    def test_trivial_read(self):
+        tree = self.make_branch_and_tree('t1')
+        self.assertIsInstance(tree.read_working_inventory(), inventory.Inventory)
+
+    def test_read_after_inventory_modification(self):
+        tree = self.make_branch_and_tree('tree')
+        # prepare for a series of changes that will modify the 
+        # inventory
+        tree.lock_write()
+        try:
+            tree.set_root_id('new-root')
+            # having dirtied the inventory, we can now expect an 
+            # InventoryModified exception when doing a read_working_inventory()
+            self.assertRaises(errors.InventoryModified, tree.read_working_inventory)
+        finally:
+            tree.unlock()

=== modified file 'bzrlib/tests/workingtree_implementations/test_set_root_id.py'
--- bzrlib/tests/workingtree_implementations/test_set_root_id.py	2006-10-04 07:42:28 +0000
+++ bzrlib/tests/workingtree_implementations/test_set_root_id.py	2006-10-05 07:34:58 +0000
@@ -38,8 +38,13 @@
         # setting the root id allows it to be read via get_root_id.
         tree.lock_write()
         try:
+            old_id = tree.get_root_id()
             tree.set_root_id(u'\xe5n id')
             self.assertEqual(u'\xe5n id', tree.get_root_id())
+            # set root id should not have triggered a flush of the tree,
+            # so check a new tree sees the old state.
+            reference_tree = tree.bzrdir.open_workingtree()
+            self.assertEqual(old_id, reference_tree.get_root_id())
         finally:
             tree.unlock()
         # having unlocked the tree, the value should have been 

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2006-10-04 07:42:28 +0000
+++ bzrlib/workingtree.py	2006-10-05 07:34:58 +0000
@@ -260,7 +260,7 @@
             self.basedir = wt.basedir
             self._control_files = wt._control_files
             self._hashcache = wt._hashcache
-            self._set_inventory(wt._inventory)
+            self._set_inventory(wt._inventory, dirty=False)
             self._format = wt._format
             self.bzrdir = wt.bzrdir
         from bzrlib.hashcache import HashCache
@@ -308,9 +308,14 @@
             hc.write()
 
         if _inventory is None:
-            self._set_inventory(self.read_working_inventory())
+            self._inventory_is_modified = False
+            self.read_working_inventory()
         else:
-            self._set_inventory(_inventory)
+            # the caller of __init__ has provided an inventory,
+            # we assume they know what they are doing - as its only
+            # the Format factory and creation methods that are
+            # permitted to do this.
+            self._set_inventory(_inventory, dirty=False)
 
     branch = property(
         fget=lambda self: self._branch,
@@ -331,9 +336,19 @@
         self._control_files.break_lock()
         self.branch.break_lock()
 
-    def _set_inventory(self, inv):
+    def _set_inventory(self, inv, dirty):
+        """Set the internal cached inventory.
+
+        :param inv: The inventory to set.
+        :param dirty: A boolean indicating whether the inventory is the same
+            logical inventory as whats on disk. If True the inventory is not
+            the same and should be written to disk or data will be lost, if
+            False then the inventory is the same as that on disk and any
+            serialisation would be unneeded overhead.
+        """
         assert inv.root is not None
         self._inventory = inv
+        self._inventory_is_modified = dirty
 
     @staticmethod
     def open(path=None, _unsupported=False):
@@ -850,6 +865,17 @@
         else:
             return '?'
 
+    def flush(self):
+        """Write the in memory inventory to disk."""
+        # TODO: Maybe this should only write on dirty ?
+        if self._control_files._lock_mode != 'w':
+            raise errors.NotWriteLocked(self)
+        sio = StringIO()
+        bzrlib.xml5.serializer_v5.write_inventory(self._inventory, sio)
+        sio.seek(0)
+        self._control_files.put('inventory', sio)
+        self._inventory_is_modified = False
+
     def list_files(self, include_root=False):
         """Recursively list all files as (path, class, kind, id, entry).
 
@@ -1011,7 +1037,10 @@
         # create a file in this interval and then the rename might be
         # left half-done.  But we should have caught most problems.
         orig_inv = deepcopy(self.inventory)
+        original_modified = self._inventory_is_modified
         try:
+            if len(from_paths):
+                self._inventory_is_modified = True
             for f in from_paths:
                 name_tail = splitpath(f)[-1]
                 dest_path = pathjoin(to_name, name_tail)
@@ -1025,7 +1054,7 @@
                             ["rename rolled back"])
         except:
             # restore the inventory on error
-            self._set_inventory(orig_inv)
+            self._set_inventory(orig_inv, dirty=original_modified)
             raise
         self._write_inventory(inv)
         return result
@@ -1471,12 +1500,21 @@
         
     @needs_read_lock
     def read_working_inventory(self):
-        """Read the working inventory."""
+        """Read the working inventory.
+        
+        :raises errors.InventoryModified: When the current in memory
+            inventory has been modified, read_working_inventory will
+            fail.
+        """
+        # conceptually this should be an implementation detail of the tree. 
+        # XXX: Deprecate this.
         # ElementTree does its own conversion from UTF-8, so open in
         # binary.
+        if self._inventory_is_modified:
+            raise errors.InventoryModified(self)
         result = bzrlib.xml5.serializer_v5.read_inventory(
             self._control_files.get('inventory'))
-        self._set_inventory(result)
+        self._set_inventory(result, dirty=False)
         return result
 
     @needs_tree_write_lock
@@ -1571,14 +1609,22 @@
             file_id = ROOT_ID
         inv = self._inventory
         orig_root_id = inv.root.file_id
+        # TODO: it might be nice to exit early if there was nothing
+        # to do, saving us from trigger a sync on unlock.
+        self._inventory_is_modified = True
+        # we preserve the root inventory entry object, but
+        # unlinkit from the byid index
         del inv._byid[inv.root.file_id]
         inv.root.file_id = file_id
+        # and link it into the index with the new changed id.
         inv._byid[inv.root.file_id] = inv.root
+        # and finally update all children to reference the new id.
+        # XXX: this should be safe to just look at the root.children
+        # list, not the WHOLE INVENTORY.
         for fid in inv:
             entry = inv[fid]
             if entry.parent_id == orig_root_id:
                 entry.parent_id = inv.root.file_id
-        self._write_inventory(inv)
 
     def unlock(self):
         """See Branch.unlock.
@@ -1681,12 +1727,8 @@
     @needs_tree_write_lock
     def _write_inventory(self, inv):
         """Write inventory as the current inventory."""
-        sio = StringIO()
-        bzrlib.xml5.serializer_v5.write_inventory(inv, sio)
-        sio.seek(0)
-        self._control_files.put('inventory', sio)
-        self._set_inventory(inv)
-        mutter('wrote working inventory')
+        self._set_inventory(inv, dirty=True)
+        self.flush()
 
     def set_conflicts(self, arg):
         raise UnsupportedOperation(self.set_conflicts, self)
@@ -1743,8 +1785,12 @@
 
     def unlock(self):
         # we share control files:
-        if self._hashcache.needs_write and self._control_files._lock_count==3:
-            self._hashcache.write()
+        if self._control_files._lock_count == 3:
+            # _inventory_is_modified is always False during a read lock.
+            if self._inventory_is_modified:
+                self.flush()
+            if self._hashcache.needs_write:
+                self._hashcache.write()
         # reverse order of locking.
         try:
             return self._control_files.unlock()
@@ -1808,8 +1854,12 @@
         return ConflictList.from_stanzas(RioReader(confile))
 
     def unlock(self):
-        if self._hashcache.needs_write and self._control_files._lock_count==1:
-            self._hashcache.write()
+        if self._control_files._lock_count == 1:
+            # _inventory_is_modified is always False during a read lock.
+            if self._inventory_is_modified:
+                self.flush()
+            if self._hashcache.needs_write:
+                self._hashcache.write()
         # reverse order of locking.
         try:
             return self._control_files.unlock()
@@ -1961,7 +2011,6 @@
                          _internal=True,
                          _format=self,
                          _bzrdir=a_bzrdir)
-        wt._write_inventory(inv)
         wt.set_root_id(inv.root.file_id)
         basis_tree = branch.repository.revision_tree(revision)
         wt.set_parent_trees([(revision, basis_tree)])
@@ -2033,7 +2082,7 @@
         branch = a_bzrdir.open_branch()
         if revision_id is None:
             revision_id = branch.last_revision()
-        inv = Inventory() 
+        inv = Inventory()
         wt = WorkingTree3(a_bzrdir.root_transport.local_abspath('.'),
                          branch,
                          inv,
@@ -2043,7 +2092,7 @@
                          _control_files=control_files)
         wt.lock_tree_write()
         try:
-            wt._write_inventory(inv)
+            # set_root_id will write the inventory to disk.
             wt.set_root_id(inv.root.file_id)
             basis_tree = branch.repository.revision_tree(revision_id)
             if revision_id == bzrlib.revision.NULL_REVISION:

